page,132
;EM	CAS1.ASM - CAS API Mapper Using DPMI
;
;	Copyright 1992, Cherry Hill Software
;
;	SUMMARY
;	    This mapper takes CAS calls issued from the Windows System
;	    VM in protected mode and redirects them to the System VM
;	    in real (V86) mode using DPMI
;
;	NOTES
;	    This mapper only supports Windows applications in standard
;	    or enhanced mode.  It does not support CAS mapping for other
;	    VMs in 386 enhanced mode.
;

.xlist
DOS5=1	    ; suppress real-mode windows support
include cmacros.inc
include segs.inc
include ..\include\dpmi.h
.list
?WIN=0	    ; Do not Use Windows entry/exit code

EXTRN DOS3CALL:FAR,GLOBALDOSALLOC:FAR,GLOBALDOSFREE:FAR

Words STRUC
LoWord dw ?
HiWord dw ?
Words ENDS

MULTIPLEX EQU 2Fh
CAS_MULTIPLEX EQU 0CBh

Int21 macro
 call DOS3CALL		; more direct than INT 21h in protected mode
 endm

sBegin Data

lpfnPrevTrapServiceRoutine dd 0
lpDosMem dd 0

regs RMCS <>		; DPMI real mode call structure
			; It's OK to be static, since we're not
			;  re-entrant

sEnd Data


sBegin	INIT_TEXT
segData dw DGroup
assumes CS,INIT_TEXT
assumes DS,DGroup


;EP	HOOKVECTOR - Hook the Multiplex Trap Vector
;
;	PROTOTYPE
;	    VOID FAR PASCAL HookVector(void) ;
;
;	ENTRY
;	    None
;
;	EXIT
;	    None
;
;	COMMENTS
;	    We save the prior vector for chaining and then insert
;	    our own.  This function is in the INIT_TEXT segment so
;	    that it may be discarded after initialization.
;
;	WARNINGS
;	    We provide no method of unhooking the vector.  This DLL
;	    must not be unloaded before unloading Windows.
;

PUBLIC HOOKVECTOR
HOOKVECTOR PROC FAR

	push	ds
	mov	ds,segData

	mov	al,MULTIPLEX
	mov	ah,35h
	Int21
	mov	lpfnPrevTrapServiceRoutine.LoWord,bx	; save prior value
	mov	lpfnPrevTrapServiceRoutine.HiWord,es	; (for chaining)

	mov	ax,_TEXT
	mov	ds,ax
	assumes DS,NOTHING
	mov	dx,OFFSET _TEXT:TrapServiceRoutine
	mov	ah,25h
	mov	al,MULTIPLEX
	Int21			    ; put in our address
	pop	ds
	ret

HOOKVECTOR endp


sEnd INIT_TEXT



page
sBegin	Code
segData2 dw DGroup
assumes CS,Code
assumes DS,DGroup

;LP	xstrlen - private strlen() function
;
;	ENTRY
;	    BP:DX = ASCIIZ string
;
;	EXIT
;	    AX = string length
;
;	COMMENTS
;
;
;	WARNINGS
;	    We provide no method of unhooking the vector.  This DLL
;	    must not be unloaded before unloading Windows.
;
;	REGISTERS USED
;	    AX,ES,flags; All others retained
;
xstrlen proc near	    ; BP:DX = ASCIIZ string
	push	ds
	mov	ds,bp
	assumes DS,nothing
	push	cx
	push	di

	push	ds
	pop	es
	mov	di,dx	    ; ES:DI = ASCIIZ string

	mov	cx,-1	    ; Search forever
	xor	ax,ax	    ; for a null byte
	repne scasb	    ; stop when matches
	not	cx
	dec	cx	    ; 0xFFFF if not found
	mov	ax,cx

	pop	di
	pop	cx
	pop	ds
	assumes ds,DGroup
	ret		    ; AX = string length
xstrlen endp


;LP	InitRMCS - Initialize the Real-Mode Call Structure
;
;	ENTRY
;	    Registers AX, BX, CX, DX, SI, and DI to copy to RMCS
;
;	EXIT
;	    RMCS initialized
;
;	SUMMARY
;	    We initialize the RMCS to zeros and fill in the values
;	    for AX, BX, CX, DX, SI, and DI as given.
;
;	WARNINGS
;	    This routine does not set up or translate the segment
;	    registers.
;
;	REGISTERS USED
;	    ES, flags; All others retained
;
InitRMCS proc near
	push	ax
	push	cx
	push	di

	mov	cx,(size RMCS)/2	; size is even
	xor	ax,ax
	push	ds
	pop	es
	mov	di,OFFSET DGroup:regs
	rep stosw		    ; Initialize to zeros
	pop	di
	pop	cx
	pop	ax

	mov	regs.rmcs_ax,ax
	mov	regs.rmcs_bx,bx
	mov	regs.rmcs_cx,cx
	mov	regs.rmcs_dx,dx
	mov	regs.rmcs_si,si
	mov	regs.rmcs_di,di
	mov	regs.rmcs_flags,7202h

	ret

InitRMCS endp


;LP	SimulateInt - Simulate Real Mode Interrupt
;
;	ENTRY
;	    The RMCS structure must be already initialized.
;
;	EXIT
;	    CY = 0
;		Registers AX, BX, CX, DX, SI, and DI as from real-mode.
;	    CY = 1
;		As above, or if DPMI error, AX = DPMI error code.
;
;	SUMMARY
;	    We call DPMI to simulate the interrupt and copy the
;	    structure to the return registers.
;	    for AX, BX, CX, DX, SI, and DI as given.
;
;	WARNINGS
;	    This routine does not translate the segment registers.
;	    The return value is indeterminate if CY is set.  Thus,
;	    this routine is most useful for functions that return
;	    with carry clear.
;
;	REGISTERS USED
;	    AX, BX, CX, DX, SI, DI, ES, flags; All others retained
;
SimulateInt proc near
	mov	ax,DPMI_SIMRMINT
	mov	bx,MULTIPLEX
	xor	cx,cx
	push	ds
	pop	es
	mov	di,OFFSET DGroup:regs
	int	IVEC_DPMI
	jc	sint1		    ; Skip if DPMI error

	push	regs.rmcs_flags
	popf
	mov	ax,regs.rmcs_ax
	mov	bx,regs.rmcs_bx
	mov	cx,regs.rmcs_cx
	mov	dx,regs.rmcs_dx
	mov	si,regs.rmcs_si
	mov	di,regs.rmcs_di
sint1:
	ret

SimulateInt endp


;LP	NoTranslation - Perform no translation for the Function Mapping
;
;	ENTRY
;	    AX, BX, CX, DX, SI, DI to pass to real mode.
;
;	EXIT
;	    CY = 0
;		Registers AX, BX, CX, DX, SI, and DI as from real-mode.
;	    CY = 1
;		As above, or if DPMI error, AX = DPMI error code.
;
;	SUMMARY
;	    We call DPMI to simulate the interrupt and copy the
;	    structure to the return registers.
;	    for AX, BX, CX, DX, SI, and DI as given.
;
;	WARNINGS
;	    This routine does not translate the segment registers.
;	    The return value is indeterminate if CY is set.  Thus,
;	    this routine is most useful for functions that return
;	    with carry clear.
;
;	REGISTERS USED
;	    AX, BX, CX, DX, SI, DI, ES, flags; All others retained
;
NoTranslation proc near
	call	InitRMCS
	call	SimulateInt
	ret
NoTranslation endp


;LP	Failure - Indicate CAS function Failure
;
;	ENTRY
;
;	EXIT
;	    CY = 0
;	    AX = CAS ERROR: Bad Function Number
;
;	SUMMARY
;
;	WARNINGS
;
;	NOTES
;	    The return value is the negative error code.
;
;	REGISTERS USED
;	    AX; All others retained
;
Failure proc near
	mov	ax,-201h	; Error: Unknown command (bad function number)
	ret
Failure endp


;LP	CantUnload - Fail the CAS Unload Function
;
;	ENTRY
;
;	EXIT
;	    CY = 0
;	    AX = CAS ERROR: Can't uninstall the Resident Manager
;
;	SUMMARY
;
;	WARNINGS
;
;	NOTES
;	    The return value is the negative error code.
;
;	REGISTERS USED
;	    AX; All others retained
;
CantUnload proc near
	mov	ax,-20Bh	; Error: Can't uninstall the Resident Manager
	ret
CantUnload endp


;LP	TranslateASCIIZDown - Copy an ASCIIZ String to Real-Mode Memory
;
;	ENTRY
;	    BP:DX - Address of String to copy
;	    AX, BX, CX, DI to pass to real mode.
;
;	EXIT
;	    CY = 0
;		Registers AX, BX, CX, DX, SI, and DI as from real-mode.
;	    CY = 1
;		As above, or if DPMI error, AX = DPMI error code.
;	    DX returned unmodified
;
;	SUMMARY
;	    We copy the string down and set the real mode DS:DX registers
;	    to the string and call SimulateInt.
;
;	WARNINGS
;	    The return value is indeterminate if CY is set.  Thus,
;	    this routine is most useful for functions that return
;	    with carry clear.
;
;	    SI is not copied to real mode.
;
;	REGISTERS USED
;	    AX, BX, CX, SI, DI, ES, flags; All others retained
;
TranslateASCIIZDown proc near
	mov	si,ax		; save ax
	call	xstrlen 	; get the string length
	inc	ax		; add one for the NUL byte
	xchg	si,ax		; restore ax, length to si
	jmp	short TranslateDown
TranslateASCIIZDown endp



;LP	TranslateSFTRDown - Copy an SFTR Structure to Real-Mode Memory
;
;	ENTRY
;	    BP:DX - Address of structure to copy
;	    AX, BX, CX, DI to pass to real mode.
;
;	EXIT
;	    CY = 0
;		Registers AX, BX, CX, DX, SI, and DI as from real-mode.
;	    CY = 1
;		As above, or if DPMI error, AX = DPMI error code.
;	    DX returned unmodified
;
;	SUMMARY
;	    We copy the structure down and set the real mode DS:DX registers
;	    to the structure and call SimulateInt.
;
;	WARNINGS
;	    The return value is indeterminate if CY is set.  Thus,
;	    this routine is most useful for functions that return
;	    with carry clear.
;
;	    SI is not copied to real mode.
;
;	REGISTERS USED
;	    AX, BX, CX, SI, DI, ES, flags; All others retained
;
TranslateSFTRDown proc near
	push	dx
	xor	si,si
	push	bx
	mov	dx,bx
	cmp	byte ptr [bx+230],1	 ; Cover page text present?
	pop	bx
	jne	tsd1		; skip if not
	mov	si,ax
	add	dx,254
	call	xstrlen 	; get the string length
	inc	ax		; add one for the NUL byte
	xchg	si,ax
tsd1:
	add	si,254		; add record length
	pop	dx
	; jmp	  short TranslateDown
TranslateSFTRDown endp


;LP	TranslateDown - Copy a Buffer to Real-Mode Memory
;
;	ENTRY
;	    BP:DX - Address of buffer to copy
;	    SI = size, in bytes, of buffer.
;	    AX, BX, CX, DI to pass to real mode.
;
;	EXIT
;	    CY = 0
;		Registers AX, BX, CX, DX, SI, and DI as from real-mode.
;	    CY = 1
;		As above, or if DPMI error, AX = DPMI error code.
;	    DX returned unmodified
;
;	SUMMARY
;	    We copy the buffer down and set the real mode DS:DX registers
;	    to the buffer and call SimulateInt.
;
;	WARNINGS
;	    The return value is indeterminate if CY is set.  Thus,
;	    this routine is most useful for functions that return
;	    with carry clear.
;
;	    SI is not copied to real mode.
;
;	REGISTERS USED
;	    AX, BX, CX, SI, DI, ES, flags; All others retained
;
TranslateDown proc near
	call	InitRMCS

	push	dx
	push	0
	push	si
	call	GLOBALDOSALLOC	; Allocate the translation buffer
	mov	lpDosMem.HiWord,ax
	mov	regs.rmcs_ds,dx
	mov	regs.rmcs_dx,0
	pop	dx
	or	ax,ax
	jz	NotEnoughMemory

	mov	cx,si
	les	di,lpDosMem
	push	ds
	mov	ds,bp
	assumes ds,nothing
	mov	si,dx
	rep movsb		; copy the data to low memory
	pop	ds
	assumes ds,DGroup

	push	dx
	call	SimulateInt	; call CAS
	pop	dx

	pushf
	mpush	?ax+?bx+?cx+?dx
	push	lpDosMem.HiWord
	call	GLOBALDOSFREE	    ; release the translation buffer
	mpop	?ax+?bx+?cx+?dx
	popf

	ret

TranslateDown endp


;LP	NotEnoughMemory - Indicate CAS function Failure
;
;	ENTRY
;
;	EXIT
;	    CY = 0
;	    AX = CAS ERROR: Not Enough Memory
;
;	SUMMARY
;
;	WARNINGS
;
;	NOTES
;	    The return value is the negative error code.
;
;	REGISTERS USED
;	    AX; All others retained
;
NotEnoughMemory proc near
	mov	ax,-0308h
	ret
NotEnoughMemory endp

;LP	Translate128Up - Simulate Interrupt and Copy 256-byte Result Up
;
;	ENTRY
;	    BP:DX - Address of return buffer
;	    AX, BX, CX, DI to pass to real mode.
;
;	EXIT
;	    CY = 0
;		Registers AX, BX, CX, DX, SI, and DI as from real-mode.
;		Returned buffer from real mode.
;	    CY = 1
;		As above, or if DPMI error, AX = DPMI error code.
;	    DX returned unmodified
;
;	SUMMARY
;	    We call SimulateInt, translating DS:DX to a real mode buffer,
;	    then copy the data to the high-memory buffer.
;
;	WARNINGS
;	    The return value is indeterminate if CY is set.  Thus,
;	    this routine is most useful for functions that return
;	    with carry clear.
;
;	    SI is not copied to real mode.
;
;	REGISTERS USED
;	    AX, BX, CX, SI, DI, ES, flags; All others retained
;
Translate128Up proc near
	mov	si,128
	jmp	short TranslateUp
Translate128Up endp

;LP	Translate256Up - Simulate Interrupt and Copy 256-byte Result Up
;
;	ENTRY
;	    BP:DX - Address of return buffer
;	    AX, BX, CX, DI to pass to real mode.
;
;	EXIT
;	    CY = 0
;		Registers AX, BX, CX, DX, SI, and DI as from real-mode.
;		Returned buffer from real mode.
;	    CY = 1
;		As above, or if DPMI error, AX = DPMI error code.
;	    DX returned unmodified
;
;	SUMMARY
;	    We call SimulateInt, translating DS:DX to a real mode buffer,
;	    then copy the data to the high-memory buffer.
;
;	WARNINGS
;	    The return value is indeterminate if CY is set.  Thus,
;	    this routine is most useful for functions that return
;	    with carry clear.
;
;	    SI is not copied to real mode.
;
;	REGISTERS USED
;	    AX, BX, CX, SI, DI, ES, flags; All others retained
;
Translate256Up proc near
	mov	si,256
	jmp	short TranslateUp
Translate256Up endp

;LP	Translate512Up - Simulate Interrupt and Copy 512-byte Result Up
;
;	ENTRY
;	    BP:DX - Address of return buffer
;	    AX, BX, CX, DI to pass to real mode.
;
;	EXIT
;	    CY = 0
;		Registers AX, BX, CX, DX, SI, and DI as from real-mode.
;		Returned buffer from real mode.
;	    CY = 1
;		As above, or if DPMI error, AX = DPMI error code.
;	    DX returned unmodified
;
;	SUMMARY
;	    We call SimulateInt, translating DS:DX to a real mode buffer,
;	    then copy the data to the high-memory buffer.
;
;	WARNINGS
;	    The return value is indeterminate if CY is set.  Thus,
;	    this routine is most useful for functions that return
;	    with carry clear.
;
;	    SI is not copied to real mode.
;
;	REGISTERS USED
;	    AX, BX, CX, SI, DI, ES, flags; All others retained
;
Translate512Up proc near
	mov	si,512
	; jmp	  TranslateUp
Translate512Up endp


;LP	TranslateUp - Simulate Interrupt and Copy Result Up
;
;	ENTRY
;	    BP:DX - Address of return buffer
;	    SI = size, in bytes, of buffer (**must be even**)
;	    AX, BX, CX, DI to pass to real mode.
;
;	EXIT
;	    CY = 0
;		Registers AX, BX, CX, DX, SI, and DI as from real-mode.
;		Returned buffer from real mode.
;	    CY = 1
;		As above, or if DPMI error, AX = DPMI error code.
;	    DX returned unmodified
;
;	SUMMARY
;	    We call SimulateInt, translating DS:DX to a real mode buffer,
;	    then copy the data to the high-memory buffer.
;
;	WARNINGS
;	    The return value is indeterminate if CY is set.  Thus,
;	    this routine is most useful for functions that return
;	    with carry clear.
;
;	    SI is not copied to real mode.
;
;	REGISTERS USED
;	    AX, BX, CX, SI, DI, ES, flags; All others retained
;
TranslateUp proc near
	call	InitRMCS

	push	dx
	push	0
	push	si
	call	GLOBALDOSALLOC	; Allocate the translation buffer
	mov	lpDosMem.HiWord,ax
	mov	regs.rmcs_ds,dx
	mov	regs.rmcs_dx,0
	pop	dx
	or	ax,ax
	jz	NotEnoughMemory

	mpush	?si+?dx
	call	SimulateInt
	mpop	?si+?dx

	pushf
	mpush	?ax+?bx+?cx+?dx

	mov	cx,si
	mov	es,bp
	mov	di,dx
	push	ds
	lds	si,lpDosMem
	assumes ds,nothing
	shr cx,1
	rep movsw
	pop	ds
	assumes ds,Data

	push	lpDosMem.HiWord
	call	GLOBALDOSFREE	    ; release the translation buffer
	mpop	?ax+?bx+?cx+?dx
	popf
	ret

TranslateUp endp

page
;
;   CAS Function Service Vector
;

ServiceVector label word
	dw	NoTranslation	    ; 00 - Get Installed State
	dw	TranslateASCIIZDown ; 01 - Submit a Task
	dw	NoTranslation	    ; 02 - Abort the Current Event
	dw	Failure 	    ; 03 - Reserved
	dw	Failure 	    ; 04 - Reserved
	dw	NoTranslation	    ; 05 - Find First Entry in Queue
	dw	NoTranslation	    ; 06 - Find Next Entry in Queue
	dw	NoTranslation	    ; 07 - Open a File
	dw	NoTranslation	    ; 08 - Delete a File
	dw	NoTranslation	    ; 09 - Delete All Files (in a queue)
	dw	NoTranslation	    ; 0A - Get Event Date
	dw	NoTranslation	    ; 0B - Set Task Date
	dw	NoTranslation	    ; 0C - Get Event Time
	dw	NoTranslation	    ; 0D - Set Task Time
	dw	Translate256Up	    ; 0E - Get External Data Block
	dw	NoTranslation	    ; 0F - Get/Set Autoreceive State
	dw	Translate512Up	    ; 10 - Get Current Event Status
	dw	NoTranslation	    ; 11 - Get Queue Status
	dw	Translate128Up	    ; 12 - Get Hardware Status
	dw	NoTranslation	    ; 13 - Run Diagnostics
	dw	TranslateASCIIZDown ; 14 - Move Received File
	dw	TranslateSFTRDown   ; 15 - Submit a Single File to Send
	dw	CantUnload	    ; 16 - Unload Resident Manager
	dw	NoTranslation	    ; 17 - Set Cover Page Status

NUMSERVICES equ (($-ServiceVector)/2)

page

assumes DS,nothing

;LP	TrapServiceRoutine - Multiplex Interrupt Trap Service Routine
;
;	ENTRY
;	    AX = Function Code
;
;	EXIT
;	    Jump to prior routine if not our function code, else:
;
;	    Vector to CAS function.
;	    CY = 0
;		Registers AX, BX, CX, DX, SI, and DI as from real-mode.
;	    CY = 1
;		As above, or if DPMI error, AX = DPMI error code.
;
;	SUMMARY
;	    Once we determine that the function is ours, we copy the
;	    caller's DS register to BP and vector to the mapping routine.
;
;	WARNINGS
;	    The return value is indeterminate if CY is set.  Thus,
;	    this routine is most useful for functions that return
;	    with carry clear.
;
;	    SI is not copied to real mode.
;
;	REGISTERS USED
;	    AX, BX, CX, DX returned per CAS function.
;	    SI, DI, BP, DS, and ES retained.
;
TrapServiceRoutine proc far
	sti
	cmp	ah,CAS_MULTIPLEX
	je	tsr1

	push	0
	push	0		; push empty CALL frame
	push	bp
	mov	bp,sp
	push	ds
	mov	ds,segData2
	assumes ds,Data
	push	ax
	mov	ax,lpfnPrevTrapServiceRoutine.LoWord
	mov	[bp+2].LoWord,ax
	mov	ax,lpfnPrevTrapServiceRoutine.HiWord
	mov	[bp+2].HiWord,ax
	pop	ax
	pop	ds
	assumes ds,nothing
	leave
	retf
tsr1:
	cmp	al,NUMSERVICES
	jae	tsr3		    ; We don't go that high

	push	bp
	mov	bp,sp
	push	si
	push	di
	push	ds
	push	es
	push	bp
	push	OFFSET _TEXT:tsr2   ; push CALL frame to return back here
	push	0FFFFh		    ; push CALL frame that we'll call
	push	bp
	mov	bp,sp
	push	bx
	mov	bx,ax
	xor	bh,bh
	add	bx,bx
	mov	bx,ServiceVector[bx]
	mov	[bp+2],bx	    ; store actual return address
	pop	bx
	leave
	mov	bp,ds
	mov	ds,segData2
	retn			; call service routine
tsr2:				; resume here after processing
	pop	bp
	pop	es
	pop	ds
	pop	di
	pop	si
	leave
	cld			; This must always remain clear
	ret 2			; retain flags on return

; exit here with unsupported function code
;  (we don't know what, if any, translation is needed)
tsr3:
	call	Failure
	cld			; This must always remain clear
	ret 2			; retain flags on return
TrapServiceRoutine endp

sEnd Code

end
