/*EM	TIMER.C - Timer Device Driver DLL
 *
 *	SUMMARY
 *	    Basic LibMain and WEP functions
 *
 *	COMMENTS
 *
 *	WARNINGS
 *
 */


#include <windows.h>
#include "timera.h"
#include "pic.h"
#include "dpmi.h"

#define EXPORT _export _loadds
#include "chtimer.h"

#define TIMER_IRQ	 0

/* Set variables for our interrupt number */

#define INT_DEV (INT_MASTER_0+(TIMER_IRQ & 7))
#define PIC00 INTA00
#define PIC01 INTA01
#define INT_MASK (1 << (TIMER_IRQ & 7))


BOOL FAR PASCAL LibMain(HANDLE hInstance   /* Library instance handle  */
		       ,WORD wDataSeg	    /* Default data segment	*/
		       ,WORD cbHeap	    /* Default Heap size	*/
		       ,LPSTR lpszCmdLine) ;/* Command line		*/
int FAR PASCAL WEP(int fSystemExit) ;
#pragma alloc_text(INIT_TEXT,LibMain)	/* Keep this with LIBENTRY.ASM	*/
#pragma alloc_text(FIXED_TEXT,WEP)

HANDLE hLibInstance ;
HANDLE hReflector ;

FARPROC GetPMVector(int iVector)
{
    FARPROC dwRet ;

    _asm {
	mov	al,byte ptr iVector
	mov	ah,35h
	int	21h
	mov	word ptr dwRet,bx
	mov	word ptr dwRet+2,es ; Save it
	}
    return dwRet ;
}


void SetPMVector(int iVector, FARPROC lpfnISR)
{
    _asm {
	push	ds
	lds	dx,lpfnISR
	mov	al,byte ptr iVector
	mov	ah,25h
	int	21h		    ; Set our ISR
	pop	ds
	}
}


/*XP<	LibMain - Main library entry point.
 *
 *	ENTRY
 *
 *	EXIT
 *
 *	RETURNS
 *	    TRUE if initialization succeeds, FALSE otherwise.
 *
 *	WARNINGS
 *
 *	CALLS
 *
 *	NOTES
 *	    The true library entry point is in the assembly language
 *	    module LIBENTRY.ASM, but it simply passes control to here.
 *
 */
BOOL FAR PASCAL LibMain(HANDLE hInstance   /* Library instance handle  */
		       ,WORD wDataSeg	    /* Default data segment	*/
		       ,WORD cbHeap	    /* Default Heap size	*/
		       ,LPSTR lpszCmdLine)  /* Command line		*/
/*>*/
{
    lpszCmdLine = lpszCmdLine ;     /* Avoid -W4 warnings		*/
    wDataSeg = wDataSeg ;
    cbHeap = cbHeap ;
    hLibInstance = hInstance ;	    /* We may need this later to access
				       resources from our executable	*/

    return TRUE ;
}


/*XP<	WEP - Windows Exit Procedure
 *
 *	ENTRY
 *	    fSystemExit indicates if we are terminating the Windows
 *	    session.  Otherwise, we are only unloading this DLL.
 *
 *	RETURNS
 *	    Always returns "1".
 *
 *	WARNINGS
 *	    Due to bugs in Windows 3.0 and earlier (and perhaps later),
 *	    this function must be placed in a fixed segment.  This same
 *	    bug makes the value of DS questionable, so we cannot use it
 *	    (or any static data).
 *
 *	    To be sure, we don't do anything here, anyway.
 *
 *	CALLS
 *	    None
 *
 *	NOTES
 *	    This is the standard DLL exit procedure.
 *
 */
int FAR PASCAL WEP(int fSystemExit)
/*>*/
{
    fSystemExit = fSystemExit ;     /* Avoid -W4 warning		*/
    return 1 ;			    /* Always indicate success		*/
}


void EXPORT FAR PASCAL TimerStart(HWND hWnd, WPARAM wParam)
{
    wParamEvent = wParam ;
    hWndEvent = hWnd ;

    if (!lpfnPrevISR)
	{
	/* Save the previous ISR and set ours */
	_asm cli
	lpfnPrevISR = GetPMVector(INT_DEV) ;
	SetPMVector(INT_DEV,(FARPROC)IntSvcRtn) ;

	/* Unmask the interrupt and begin device I/O */
	_asm sti
	}
}


int EXPORT FAR PASCAL TimerGetCount(void)
{
    WORD wCountRet ;

    _asm {
	mov	ax,SEG wCount
	mov	es,ax
	xor	ax,ax
	xchg	ax,es:wCount		; get count, set to zero
	mov	wCountRet,ax
	}

    return wCountRet ;
}


void EXPORT FAR PASCAL TimerStop(void)
{
    hWndEvent = 0x0000 ;	/* Tell the ISR to stop */

    if (!lpfnPrevISR)
	return ;		/* return if not started */

    SetPMVector(INT_DEV, lpfnPrevISR) ; 	 /* Restore the prot-mode vector */

    lpfnPrevISR = NULL ;
}
