PAGE ,132
title VxD2A.ASM - Example Device Driver #2a
;EM	VxD2A - Example Device Driver #2a
;
;	Copyright 1992, Cherry Hill Software
;	All rights reserved.
;
;	SUMMARY
;	    This driver simulates an I/O port.	When the I/O port FAKE_PORT
;	    (see bogus.inc) is read, the virtual device returns a value
;	    and increments it by one for the subsequent read.  When the
;	    port is written, the value is set.	The initial value is zero.
;
;	WARNINGS
;
;
	.386p


.xlist
include vmm.inc
include debug.inc
include v86mmgr.inc
include vpicd.inc
include ..\include\bogus.inc
.list

VM_Not_Executable equ VM_Not_Executeable ; acckk!

subttl	VxD Declaration/Definition
page

VxD2A_Init_Order equ VNETBIOS_Init_Order+100 ; Do this after the Virtual net
VxD2A_Device_ID equ Bogus_Device_ID


Declare_Virtual_Device VXD2, 1, 0, VxD2A_Control, VxD2A_Device_ID, \
		       VxD2A_Init_Order


VxD_DATA_SEG

bFakeData   db	0	    ; Fake I/O port data

VxD_DATA_ENDS



subttl Dispatch VxD Control
page
VxD_LOCKED_CODE_SEG

;IP	Port_IO_Callback - Process access to FAKE_PORT
;
;	ENTRY
;	    EAX - The output value (for output instructions)
;	    EBX - The handle to the current VM
;	    ECX - The type of I/O operation
;	    DS,ES - FLAT
;
;	EXIT
;	    EAX the input value (for input instructions)
;
;	WARNINGS
;
;	NOTES
;	    Note that we don't even look at the client register frame.
;
;	    We simply read and increment
;
;	CALLS
;
BeginProc Port_IO_Callback, NO_LOG
	Dispatch_Byte_IO Fall_Through,Port_Output_Callback

Port_Input_Callback:
	mov	al,bFakeData
	inc	bFakeData
	ret

Port_Output_Callback:
	mov	bFakeData,al
	ret

EndProc Port_IO_Callback


ifdef DEBUG
BeginProc VxD2A_Debug_Query
	Trace_Out   "VxD2A has no debug command support."
	clc
	ret
EndProc VxD2A_Debug_Query
endif


;
;   VxD2A_Control
;

CtlDisp macro x
 Control_Dispatch x, VxD2A_&x
 endm

Begin_Control_Dispatch VxD2A
	CtlDisp Device_Init
ifdef DEBUG
	CtlDisp Debug_Query
endif
End_Control_Dispatch VxD2A

VxD_LOCKED_CODE_ENDS



VxD_CODE_SEG
VxD_CODE_ENDS


subttl	Initialization Data
page
VxD_IDATA_SEG
; put initialization-only data here
VxD_IDATA_ENDS



subttl	VxD Initialization
page
VxD_ICODE_SEG

page
;EP	VxD2A_Device_Init - Non-critical Device Initialization
;
;	ENTRY
;	    EBP - Client frame
;	    EBX - System VM handle
;	    DS,ES - FLAT
;
;	EXIT
;	    SUCCESS
;		Carry clear
;	    FAILURE
;		Carry set
;
;	WARNINGS
;
;	NOTES
;
;	CALLS
;

BeginProc VxD2A_Device_Init
	Debug_Out "VxD2A_Device_Init"
	mov	edx,FAKE_PORT
	mov	esi,OFFSET32 Port_IO_Callback
	VMMCall Install_IO_Handler
	VMMCall Enable_Global_Trapping
	clc				; No error
	ret
EndProc VxD2A_Device_Init



VxD_ICODE_ENDS


VxD_REAL_INIT_SEG

VxD2A_Real_Init LABEL FAR ; Called before Windows enters protected mode
	mov ax,Device_Load_Ok	; Allow the VxD to load
	xor bx,bx		; No EMM Exclude pages
	xor si,si		; no instance data items
				; pass edx unmodified
	ret

VxD_REAL_INIT_ENDS

	END VxD2A_Real_Init
